import numpy as np
import pandas as pd
import os
import re
import xml.etree.cElementTree as ET
from io import StringIO
from PyQt5.QtCore import QObject, pyqtSignal

import utils
from dataprovider.exploprovider import ExploProvider
from cfatools.logreader.dataset import DatasetReader
from cfatools.logreader.pump_calib import PumpCalibReader
import cfatools.processor.iceblock as iceblock_processor
import cfatools.processor.pump as pump_processor
import cfatools.processor.encoder as encoder_processor


class IcbktransitionProvider(QObject):

    sig_transition_dataset_loaded = pyqtSignal(name="transition_dataset_loaded")
    sig_iceblock_df_updated = pyqtSignal(pd.DataFrame)
    sig_transitions_df_updated = pyqtSignal(pd.DataFrame)

    def __init__(self, explo_prvd: ExploProvider):
        super(IcbktransitionProvider, self).__init__()

        self.explo_prvd = explo_prvd
        self.dataset_reader = None
        self.iceblock_filename = ""
        self._iceblock_df = pd.DataFrame()
        self._transitions_df = pd.DataFrame()
        self.encoder_df = pd.DataFrame()
        self.conduct_df = pd.DataFrame()
        self.picarro_df = pd.DataFrame()

        self.explo_prvd.sig_dataset_loaded.connect(self.load_transition_dataset)
        self.sig_iceblock_df_updated.connect(self.update_transitions_df)

    @property
    def iceblock_df(self) -> pd.DataFrame:
        return self._iceblock_df

    @iceblock_df.setter
    def iceblock_df(self, iceblock_df):
        self._iceblock_df = iceblock_df
        self.sig_iceblock_df_updated.emit(self._iceblock_df)

    @property
    def transitions_df(self) -> pd.DataFrame:
        return self._transitions_df

    @transitions_df.setter
    def transitions_df(self, transitions_df):
        self._transitions_df = transitions_df
        self.sig_transitions_df_updated.emit(self._transitions_df)

    def load_transition_dataset(self, dataset_reader: DatasetReader) -> None:
        if "CONDUCTI-R_periodic" not in dataset_reader.get_instruments_names():
            return
        self.dataset_reader = dataset_reader
        self.iceblock_filename = os.path.join(
            self.dataset_reader.dataset_path,
            "processed",
            dataset_reader.dataset_name + "_iceblocks.csv",
        )
        self.encoder_df = self.dataset_reader.get_timeseries("ENCODER_periodic")
        self.conduct_df = self.dataset_reader.get_timeseries(
            "CONDUCTI-R_periodic"
        ).copy()
        self.picarro_df = self.dataset_reader.get_timeseries(
            "PICARRO-R_periodic"
        ).copy()
        self.sig_transition_dataset_loaded.emit()

    def build_iceblock_df(
        self, retrieve_iceblocks_from_processed_csv: bool = True
    ) -> pd.DataFrame:
        if retrieve_iceblocks_from_processed_csv and os.path.exists(
            self.iceblock_filename
        ):
            iceblock_df = pd.read_csv(
                self.iceblock_filename,
                parse_dates=["datetime_start", "datetime_end", "datetime_stacked"],
            )
        else:
            iceblock_df = self.dataset_reader.get_timeseries("ICBKCTRL_instant")
            iceblock_df = iceblock_processor.get_clean_iceblock_df(iceblock_df)
        self.iceblock_df = iceblock_df
        return iceblock_df

    def save_iceblock_df_as_csv(self):
        self.iceblock_df.to_csv(
            self.iceblock_filename, index=False, date_format="%Y-%m-%dT%H:%M:%SZ"
        )

    def update_transitions_df(self):
        transitions_df = self.iceblock_df[["id", "datetime_start"]]
        transitions_df = transitions_df.rename(
            columns={"id": "to_id", "datetime_start": "datetime"}
        )
        transitions_df["from_id"] = transitions_df["to_id"].shift()
        if self.iceblock_df.iloc[-1]["datetime_end"] is not pd.NaT:
            transitions_df = transitions_df.append(
                pd.DataFrame(
                    {
                        "from_id": max(self.iceblock_df["id"]),
                        "to_id": np.nan,
                        "datetime": self.iceblock_df.iloc[-1]["datetime_end"],
                    },
                    index=[len(transitions_df.index)],
                )
            )
        transitions_df["timestamp"] = utils.pd_time_to_epoch_ms(
            transitions_df["datetime"]
        )
        self.transitions_df = transitions_df

    def move_transition(self, transition_id: int, new_timestamp_sec: int):
        transition = self.transitions_df.iloc[transition_id]
        new_timestamp = pd.Timestamp(new_timestamp_sec * 10**9, tz="UTC")
        self.iceblock_df.loc[
            self.iceblock_df["id"] == transition["from_id"], "datetime_end"
        ] = new_timestamp
        self.iceblock_df.loc[
            self.iceblock_df["id"] == transition["to_id"], "datetime_start"
        ] = new_timestamp
        self.update_iceblocks_height()
        self.sig_iceblock_df_updated.emit(self.iceblock_df)

    def delete_transition(self, transition_id: int):
        self.iceblock_df = self.iceblock_df.drop(transition_id)
        self.iceblock_df = self._rebuild_index__(self.iceblock_df)

    def _rebuild_index__(self, iceblock_df: pd.DataFrame) -> pd.DataFrame:
        iceblock_df.reset_index(drop=True, inplace=True)
        iceblock_df["id"] = range(1, len(iceblock_df.index) + 1)
        return iceblock_df

    def update_iceblocks_height(self):
        for index, iceblock in self.iceblock_df.iterrows():
            encoder_height = self.get_encoder_height_at(iceblock["datetime_start"])
            self.iceblock_df.loc[index, "initial_height"] = encoder_height

    def get_encoder_height_at(self, timestamp: pd.Timestamp):
        return self.encoder_df.iloc[
            self.encoder_df.index.get_loc(timestamp, method="nearest")
        ]["position"]
